#include "graph_manager.h"
#include "edge_se3_prior.h"
#include "g2o/types/slam3d/vertex_se3.h"

#include "edge_2DOF_prior.h"
bool GraphManager::addExtrinsicCameraCalibrationEdge(Node* new_node, double max_time_difference)
{
    Node* old_node = (--(graph_.rbegin()))->second; 
    ROS_INFO("If appropriate, using extrinsic calibration to node %d", (int)graph_.rbegin()->second->id_);
    ROS_INFO_STREAM("Header Frames:\nOld (Vid=" << old_node->vertex_id_ << "): " << old_node->pc_col->header.frame_id << "\nNew (Vid=" << new_node->vertex_id_ << "): " << new_node->pc_col->header.frame_id );
    ROS_INFO_STREAM("Header Stamps:\nOld (Vid=" << old_node->vertex_id_ << "): " << old_node->pc_col->header.stamp    << "\nNew (Vid=" << new_node->vertex_id_ << "): " << new_node->pc_col->header.stamp );
    if(old_node->pc_col->header.frame_id != new_node->pc_col->header.frame_id 
       && std::abs((new_node->pc_col->header.stamp - old_node->pc_col->header.stamp).toSec()) < max_time_difference)
       //&& calibration_edges_.size() == 1)
    {
        g2o::VertexSE3* cam1 = dynamic_cast<g2o::VertexSE3*>(optimizer_->vertex(old_node->vertex_id_));
        g2o::VertexSE3* cam2 = NULL;
        if(new_node->vertex_id_ > 0){
          cam2 = dynamic_cast<g2o::VertexSE3*>(optimizer_->vertex(new_node->vertex_id_));
        } else {
          cam2 = new g2o::VertexSE3;
          assert(cam2 && "couldn't create new g2o::VertexSE3");
          int v_id = next_vertex_id++;
          cam2->setId(v_id);
          new_node->vertex_id_ = v_id;
          camera_vertices.insert(cam2);
          optimizer_->addVertex(cam2); 
        }
        ROS_ERROR_COND(!cam1 || !cam2, "Null pointer for previous node's vertex (vid = %d)", old_node->vertex_id_);
        calibration_edge_type* calibration_edge = NULL;
        std::string first_frame_id = graph_.begin()->second->pc_col->header.frame_id; //canonical origin of the calibration displacement
        if(old_node->pc_col->header.frame_id == first_frame_id){
          calibration_edge = new calibration_edge_type(cam1, cam2, static_cast<calibration_vertex_type*>(calibration_vertex_));
          ROS_INFO("Adding Calibration edge from vertex %d to %d", cam1->id(), cam2->id());
        } else if(new_node->pc_col->header.frame_id == first_frame_id){
          calibration_edge = new calibration_edge_type(cam2, cam1, static_cast<calibration_vertex_type*>(calibration_vertex_));
          ROS_INFO("Adding Calibration edge from vertex %d to %d", cam2->id(), cam1->id());
        } else {
          ROS_ERROR("Unforeseen frame-id combination!");
        }
        Eigen::Matrix<double, 6,6> information;
        information.setIdentity();
        calibration_edge->setInformation(information*ParameterServer::instance()->get<double>("calibration_information_factor"));
        calibration_edges_.insert(calibration_edge);

        //if(!added_node){
          //Set initial guess according to calibration
          //cam2->setEstimate(calibration_vertex_->estimate() * cam1->estimate() );
          //motion_estimate = eigenTF2QMatrix(cam2->estimate()); 
        //}

        ROS_ERROR("Adding Calibration edge for Nodes %d (vid: %d) and %d (vid: %d) and calibration (vid: %d)", old_node->id_, old_node->vertex_id_, new_node->id_, new_node->vertex_id_, calibration_vertex_->id());
        QMutexLocker locker(&optimizer_mutex_);
        optimizer_->addEdge(calibration_edge);
        return true;
    }
    else {
      ROS_INFO("No Extrinsic Edge: Timestamps too different or same frame_id.");
      return false;
    }
}

void GraphManager::odomCalibrationInitialization() {
    //automatically sets identity on construction
    odom_calibration_vertex_type* odom_calibration_vertex= new odom_calibration_vertex_type();//FIXME: memory leak, initialize, delete if exist
    odom_calibration_vertex->setId(next_vertex_id++);

    double calib_info = ParameterServer::instance()->get<double>("calibration_information_factor");
    double calib_prior = calib_info * ParameterServer::instance()->get<double>("calibration_prior_factor");

    //Reduce 6DOF to 3DOF 
    Eigen::Matrix<double, 6,1> diagonal; diagonal << 1, 1, 1, 1, 1, 1;
    Eigen::Matrix<double, 7,1> prior_se3; prior_se3 << 0, 0, 0, 0.0, 0.0, 0.0, 1.0;
    //Eigen::Matrix<double, 7,1> prior_se3; prior_se3 << 0.0, 0.0, 0.0, 0.5, -0.5, 0.5, 0.5;

    odom_calibration_vertex->setEstimateData(prior_se3.data());
    if(!ParameterServer::instance()->get<bool>("calibrate_odometry_online")){
      odom_calibration_vertex->setFixed(true);
    }  
    //Don't need prior if fixed?
    g2o::MyEdgeSE3Prior* calibration_prior_edge = new g2o::MyEdgeSE3Prior();
    calibration_prior_edge->setMeasurement(odom_calibration_vertex->estimate());
    

    //GENERIC
    calibration_prior_edge->setInformation((calib_prior*diagonal).asDiagonal());
    calibration_prior_edge->vertices()[0] = odom_calibration_vertex;
    calibration_prior_edge->setMeasurement(odom_calibration_vertex->estimate());
    calibration_edges_.insert(calibration_prior_edge);
    odom_calibration_vertex_ = odom_calibration_vertex;
}

void GraphManager::calibrationInitialization() {
    //automatically sets identity on construction
    calibration_vertex_type* calibration_vertex= new calibration_vertex_type();//FIXME: memory leak, initialize, delete if exist
    calibration_vertex->setId(next_vertex_id++);

    double calib_info = ParameterServer::instance()->get<double>("calibration_information_factor");
    double calib_prior = calib_info * ParameterServer::instance()->get<double>("calibration_prior_factor");

    //catadioptric 2DOF:
    /*
    calibration_vertex->setEstimate(g2o::Vector2d(0.0, M_PI));//initial guess
    g2o::MyEdge2DOFPrior* calibration_prior_edge = new g2o::MyEdge2DOFPrior();
    //Information matrix for prior
    Eigen::Matrix<double, 2,1> diagonal; diagonal.setOnes(); 
    */

    /*catadioptric 6DOF: 
    Eigen::Matrix<double, 6,1> diagonal; diagonal.setOnes();
    Eigen::Matrix<double, 7,1> prior_se3;
    //Calibration Data for mirror setup on Monday evening: 
    //VERTEX_SE3:QUAT 0 -0.00410193 -0.000482643 0.0862538 0.979085 0.00537571 -0.0290492 0.201295 
    if(ParameterServer::instance()->get<std::string>("topic_image_mono_split").find("split") != std::string::npos) {
      //prior_se3 << -0.00410193, -0.000482643, 0.0862538, 0.979085, 0.00537571, -0.0290492, 0.201295;
      //Tuesday Night: 0.00757327 0.023033 0.110041 0.939605 0.00198476 -0.002364 0.342248
      //Newly angled structure
      //"Factory" calibration: 0.0, 0.0, 0.0, 0.88302222, 0.00, 0.0, 0.34202014;
      //Wednesday Night: 0.0146295 0.00317551 0.11388 0.944518 -0.00553805 -0.0194281 0.327839
      //BAD: prior_se3 << 0.0146295, 0.00317551, 0.11388, 0.944518, -0.00553805, -0.0194281, 0.327839;
      //From openni_listener
      prior_se3 << -0.0110499, 0.0502848, 0.099574, 0.936359, 0.0142091, -0.00807437, 0.350664;
      //Bad but canonical
      //prior_se3 << 0,0,0,0,0,0,1;

      //Information matrix for Mirror prior
    } else { 
    //Calibration Data for 2 Camera setup on Monday evening:
    //Calibration Vertex: 0.00157616 -4.08291e-05 -0.0600107 4.88448e-05 1 2.30876e-05 1.76034e-05
      prior_se3 << 0.00157616, -4.08291e-05, -0.0600107, 4.88448e-05, 1, 2.30876e-05, 1.76034e-05;
      //Information matrix for 2 Camera prior
      //diagonal << 1, calib_info, 1, calib_info, calib_info, calib_info;
    }

    calibration_vertex->setEstimateData(prior_se3.data());
    if(!ParameterServer::instance()->get<bool>("calibrate_viewpoints_online")){
      calibration_vertex->setFixed(true);
    }  
    //Don't need prior if fixed?
    g2o::MyEdgeSE3Prior* calibration_prior_edge = new g2o::MyEdgeSE3Prior();
    calibration_prior_edge->setMeasurement(calibration_vertex->estimate());
    
    */

    //GENERIC
    //calibration_prior_edge->setInformation((calib_prior*diagonal).asDiagonal());
    //calibration_prior_edge->vertices()[0] = calibration_vertex;
    //calibration_prior_edge->setMeasurement(calibration_vertex->estimate());
    //calibration_edges_.insert(calibration_prior_edge);
    calibration_vertex_ = calibration_vertex;
}


