#include "graph_manager.h"
#include "scoped_timer.h"
#include "misc.h"
void createOdometryEdge(int id1, int id2, tf::Transform& odomTf, LoadedEdge3D& edge)
{
  double infoCoeff = ParameterServer::instance()->get<double>("odometry_information_factor");
  edge.id1 = id1;
  edge.id2 = id2;
  edge.mean = tf2G2O(odomTf);

  //Real odometry
  edge.informationMatrix = Eigen::Matrix<double, 6, 6>::Identity() * infoCoeff; //0.01m /0.01rad error
  //edge.informationMatrix = Eigen::Matrix<double, 6, 6>::Zero(); //Do not influence optimization
  /*
   //Assumption: Motion in the plane
   odom_edge.informationMatrix(0,0) = 100; //0.1m accuracy in the floor plane
   odom_edge.informationMatrix(1,1) = 100; //
   odom_edge.informationMatrix(2,2) = 1000000;//0.01m information on rotation w.r. to floor
   odom_edge.informationMatrix(3,3) = 400000000;//0.02rad information on rotation w.r. to floor
   odom_edge.informationMatrix(4,4) = 400000000;//0.02rad information on rotation w.r. to floor
   odom_edge.informationMatrix(5,5) = 1600; //0.4rad (~20°) on rotation about vertical
   */
}

void GraphManager::addOdometry(ros::Time timestamp,
         tf::TransformListener* listener) 
{
  //return if graph is empty
  if (graph_.size() < 2) return;

  //Variables Used in loop
  //look through graph to find nodes that do not have odom edges
  std::map<int, Node*>::reverse_iterator rit = graph_.rbegin();
  std::map<int, Node*>::reverse_iterator prev_rit = graph_.rbegin();
  prev_rit++;

  std::string odom_fixed_frame = ParameterServer::instance()->get<std::string>("odom_frame_name");
  std::string odom_target_frame = ParameterServer::instance()->get<std::string>("odom_target_frame_name");
  std::string odom_target_frame1;//for start's frame in loop
  std::string odom_target_frame2;//for end's frame in loop
  tf::StampedTransform deltaTransform;
  std::string error_msg;
  bool ok = false;

  for (; prev_rit != graph_.rend(); ++prev_rit, ++rit) {
    Node *start = prev_rit->second;
    Node *end = rit->second;

    //if we have already received the odometry for the node which we would like to point to
    //we know we can interpolate instead of extrapolate the odometry and insert an edge as well as the odometry to the node
    if (!start->has_odometry_edge_) {
        //Get Frames
        if(odom_target_frame.empty()){//Use data frame
          odom_target_frame1 = start->pc_col->header.frame_id;
          odom_target_frame2 = end->pc_col->header.frame_id;
        } else{//Use value from parameter
          odom_target_frame1 = odom_target_frame;
          odom_target_frame2 = odom_target_frame;
        }
        ok = listener->canTransform(odom_target_frame1, start->timestamp_, //frame and time of earlier node
                                    odom_target_frame2, end->timestamp_,   //frame and time of newer node
                                    odom_fixed_frame, &error_msg);
        if(ok){
          //from http://wiki.ros.org/tf/Tutorials/Time%20travel%20with%20tf%20%28C%2B%2B%29
          //listener.lookupTransform("/turtle2", now, "/turtle1", past, "/world", transform);
          listener->lookupTransform(odom_target_frame1, start->timestamp_, //frame and time of earlier node
                                    odom_target_frame2, end->timestamp_,   //frame and time of newer node
                                    odom_fixed_frame, deltaTransform);

          printTransform("Odometry Delta", deltaTransform);
          //ADD edge here
          LoadedEdge3D edge;
          createOdometryEdge(start->id_, end->id_, deltaTransform, edge);
          QMatrix4x4 motion_estimate =eigenTF2QMatrix(edge.mean);//not used
          addOdometryEdgeToG2O(edge, start, end, motion_estimate);
          start->has_odometry_edge_=true;
        } else {//couldn't transform
          ROS_ERROR("Cannot transform between node %d (time %d.%09d) and %d (time %d.%09d). Stated reason: %s", 
                    start->id_, start->timestamp_.sec, start->timestamp_.nsec, end->id_, end->timestamp_.sec, end->timestamp_.nsec, error_msg.c_str());
          ROS_ERROR_STREAM(listener->allFramesAsString());
        }
    } else {// Encountered node with odometry edge = true
      //break;
    }

  }
}

bool GraphManager::addOdometryEdgeToG2O(const LoadedEdge3D& edge,
                                        Node* n1, Node* n2,  
                                        QMatrix4x4& motion_estimate) //Pure output
{
    ScopedTimer s(__FUNCTION__);
    assert(n1);
    assert(n2);
    assert(n1->id_ == edge.id1);
    assert(n2->id_ == edge.id2);

    QMutexLocker locker(&optimizer_mutex_);
    g2o::VertexSE3* v1 = dynamic_cast<g2o::VertexSE3*>(optimizer_->vertex(n1->vertex_id_));
    g2o::VertexSE3* v2 = dynamic_cast<g2o::VertexSE3*>(optimizer_->vertex(n2->vertex_id_));

    if(!v1 && !v2){
      ROS_ERROR("Missing both vertices: %i, %i, cannot create edge", edge.id1, edge.id2);
      return false;
    } else if (!v1 && v2) {
      ROS_ERROR("Missing previous id. This is unexpected by the programmer");
      return false;
    } else if (!v2 && v1) {
      ROS_ERROR("Creating new id for odometry. This is unexpected by the programmer");
      return false;
    }

    g2o::VertexSE3* calib = dynamic_cast<g2o::VertexSE3*>(odom_calibration_vertex_);
    v2->setEstimate(v1->estimate() * calib->estimate() * edge.mean * calib->estimate().inverse());
    //motion_estimate = eigenTF2QMatrix(v2->estimate()); 
    //if(!ParameterServer::instance()->get<bool>("calibrate_odometry_online"))
    //{ 
    //  g2o::EdgeSE3* g2o_edge = new g2o::EdgeSE3;
    //  g2o_edge->vertices()[0] = v1;
    //  g2o_edge->vertices()[1] = v2;
    //  Eigen::Isometry3d meancopy(edge.mean); 
    //  g2o_edge->setMeasurement(meancopy);
    //  g2o_edge->setInformation(edge.informationMatrix);
    //  optimizer_->addEdge(g2o_edge);
    //  ROS_DEBUG_STREAM("Added Edge ("<< edge.id1 << "-" << edge.id2 << ") to Optimizer:\n" << edge.mean.matrix() << "\nInformation Matrix:\n" << edge.informationMatrix);
    //  odometry_edges_.insert(g2o_edge);
    //  current_match_edges_.insert(g2o_edge); //Used if all previous vertices are fixed ("pose_relative_to" == "all")
    //} else 
    {
      odom_calibration_edge_type* cal_edge = new odom_calibration_edge_type(v1, v2, calib);
      cal_edge->setMeasurement(edge.mean); 
      cal_edge->setInformation(edge.informationMatrix);
      optimizer_->addEdge(cal_edge);
      ROS_INFO_STREAM("Added Odometry Edge ("<< edge.id1 << "-" << edge.id2 << ") to Optimizer");
      odometry_edges_.insert(cal_edge);
    }

    if (ParameterServer::instance()->get<std::string>("pose_relative_to") == "inaffected") {
      v1->setFixed(false);
      v2->setFixed(false);
    }
    else if(ParameterServer::instance()->get<std::string>("pose_relative_to") == "largest_loop") {
      earliest_loop_closure_node_ = std::min(earliest_loop_closure_node_, edge.id1);
      earliest_loop_closure_node_ = std::min(earliest_loop_closure_node_, edge.id2);
    }
    return true;
}
/*
// returns true, iff node could be added to the cloud
void GraphManager::addOdometry(ros::Time timestamp,
                tf::TransformListener* listener) {
        //return if graph is empty
        if (graph_.size() < 2) {
                return;
        }

        //look through graph to find nodes that do not have odom edges
        bool insert_odometry = false;
        std::map<int, Node*>::reverse_iterator rit = graph_.rbegin();
        std::map<int, Node*>::reverse_iterator prev_rit = graph_.rbegin();
        prev_rit++;

        for (; prev_rit != graph_.rend(); ++prev_rit, ++rit) {
                Node *start = prev_rit->second;
                Node *end = rit->second;
                ROS_WARN("Timestamp of Node %d: %f", start->id_, start->timestamp_.toSec());
                ROS_WARN("Timestamp of Node %d: %f", end->id_, end->timestamp_.toSec());

                //if we have already received the odometry for the node which we would like to point to
                //we know we can interpolate instead of extrapolate the odometry and insert an edge as well as the odometry to the node
                if ((end->timestamp_ < timestamp) && (!start->has_odometry_edge_)) {

                        //get start transform from node or if not yet available from tf
                        ParameterServer* ps = ParameterServer::instance();
                        std::string odom_frame = ps->get<std::string>("odom_frame_name");
                        //std::string base_frame = ps->get<std::string>("base_frame_name");
                        tf::StampedTransform start_tf;
                        if (start->odometry_set_) {
                                start_tf = start->getOdomTransform();
                        } else {
                                ROS_WARN("1: Computing odom2points for node %d", start->id_);
                                tf::StampedTransform odom2points;
                                try {
                                        listener->lookupTransform(odom_frame, start->pc_col->header.frame_id, 
                                                        start->timestamp_, odom2points);
                                } catch (tf::TransformException ex) {
                                        ROS_ERROR("%s", ex.what());
                                        tf::StampedTransform odom2base;
                                        
                                        try {
                                            listener->lookupTransform(odom_frame, ps->get<std::string>("base_frame_name"),
                                                            start->timestamp_, odom2base);
                                            {
                                              tf::Transform base2points;
                                              ROS_WARN("Using Standard kinect /openni_camera -> /openni_rgb_optical_frame as transformation ");
                                              //emulate the transformation from kinect openni_camera frame to openni_rgb_optical_frame
                                              //base2points.setRotation(tf::createQuaternionFromRPY(-1.57,M_PI,-1.57));
                                              //base2points.setOrigin(tf::Point(0,-0.04,0));
                                              base2points.setRotation(tf::createQuaternionFromRPY(0,0,0));
                                              base2points.setOrigin(tf::Point(0,0,0));
                                              odom2points  = tf::StampedTransform(odom2base * base2points, odom2points.stamp_, odom2points.frame_id_, odom2points.child_frame_id_);
                                            }
                                        } catch (tf::TransformException ex) {
                                          ROS_ERROR("Is there no odometry at all? %s", ex.what());
                                          continue;
                                        }
                                }
                                start->setOdomTransform(odom2points);
                        }

                        //get end transform from node or if not yet available from tf
                        tf::StampedTransform odom2points;
                        ROS_WARN("2: Computing odom2points for node %d", end->id_);
                        try {
                                listener->lookupTransform(odom_frame, end->pc_col->header.frame_id,
                                                end->timestamp_, odom2points);
                        } catch (tf::TransformException ex) {
                              ROS_ERROR("%s", ex.what());
                              tf::StampedTransform odom2base;
                              try {
                                      listener->lookupTransform(odom_frame, ps->get<std::string>("base_frame_name"),
                                                      start->timestamp_, odom2base);
                                      {
                                        tf::Transform base2points;
                                        ROS_WARN("Using Standard kinect /openni_camera -> /openni_rgb_optical_frame as transformation ");
                                        //emulate the transformation from kinect openni_camera frame to openni_rgb_optical_frame
                                        //base2points.setRotation(tf::createQuaternionFromRPY(-1.57,M_PI,-1.57));
                                        //base2points.setOrigin(tf::Point(0,-0.04,0));
                                        base2points.setRotation(tf::createQuaternionFromRPY(0,0,0));
                                        base2points.setOrigin(tf::Point(0,0,0));
                                        odom2points  = tf::StampedTransform(odom2base * base2points, odom2points.stamp_, odom2points.frame_id_, odom2points.child_frame_id_);
                                        printTransform("Odom to Points", odom2points);
                                          printTransform("Odom to Base", odom2base);
                                        }
                              } catch (tf::TransformException ex) {
                                ROS_ERROR("Is there no odometry at all? %s", ex.what());
                                continue;
                              }
                        } 
                        end->setOdomTransform(odom2points);

                         //Get motion estimate from transformations
                        tf::Transform delta_transform = start->getOdomTransform().inverse() * end->getOdomTransform();
                        printTransform("Odometry Delta", delta_transform);
                        //ADD edge here
                        LoadedEdge3D edge;
                        edge.id1 = start->id_;
                        edge.id2 = end->id_;
                        edge.mean = tf2G2O(delta_transform);
                        QMatrix4x4 motion_estimate = g2o2QMatrix(edge.mean);

                        //Real odometry
                        edge.informationMatrix =
                                        Eigen::Matrix<double, 6, 6>::Identity() * 100; //0.1m /0.1rad error
                         ////Assumption: Motion in the plane
                         //odom_edge.informationMatrix(0,0) = 100; //0.1m accuracy in the floor plane
                         //odom_edge.informationMatrix(1,1) = 100; //
                         //odom_edge.informationMatrix(2,2) = 1000000;//0.01m information on rotation w.r. to floor
                         //odom_edge.informationMatrix(3,3) = 400000000;//0.02rad information on rotation w.r. to floor
                         //odom_edge.informationMatrix(4,4) = 400000000;//0.02rad information on rotation w.r. to floor
                         //odom_edge.informationMatrix(5,5) = 1600; //0.4rad (~20°) on rotation about vertical

                        addOdometryEdgeToG2O(edge, start, end, motion_estimate);
                        start->has_odometry_edge_=true;
                } else if ((start->has_odometry_edge_)
                                || (end->timestamp_ > timestamp)) { //next node has odometry edge --> we are done
                        break;
                }
        }
}
*/

