/* This file is part of RGBDSLAM.
 * 
 * RGBDSLAM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * RGBDSLAM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with RGBDSLAM.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "parameter_server.h"
//Documentation see header file
#include "pcl/ros/conversions.h"
#include <pcl/io/io.h>
//#include "pcl/common/transform.h"
#include "pcl_ros/transforms.h"
#include "openni_listener.h"
#include <cv_bridge/cv_bridge.h>
#include <opencv2/imgproc/imgproc.hpp>
#include <iostream>
#include <string>
#include <cv.h>
//#include <ctime>
#include <sensor_msgs/PointCloud2.h>
#include <Eigen/Core>
#include "node.h"
#include "misc.h"
#include <g2o/types/slam3d/isometry3d_mappings.h>
//#include <image_geometry/pinhole_camera_model.h>
//#include "pcl/ros/for_each_type.h"

//For rosbag reading
#include <rosbag/view.h>
#include <boost/foreach.hpp>


#include "scoped_timer.h"
//for comparison with ground truth from mocap and movable cameras on robots
#include <tf/transform_listener.h>

typedef message_filters::Subscriber<sensor_msgs::Image> image_sub_type;      
typedef message_filters::Subscriber<sensor_msgs::CameraInfo> cinfo_sub_type;      
typedef message_filters::Subscriber<sensor_msgs::PointCloud2> pc_sub_type;      
typedef message_filters::Subscriber<nav_msgs::Odometry> odom_sub_type;



OpenNIListener::OpenNIListener(GraphManager* graph_mgr)
: graph_mgr_(graph_mgr),
  stereo_sync_(NULL), kinect_sync_(NULL), no_cloud_sync_(NULL),
  visua_sub_(NULL), depth_sub_(NULL), cloud_sub_(NULL),
  visua_sub_split_(NULL), depth_sub_split_(NULL), cinfo_sub_split_(NULL),
  depth_mono8_img_(cv::Mat()),
  save_bag_file(false),
  pause_(ParameterServer::instance()->get<bool>("start_paused")),
  getOneFrame_(false),
  first_frame_(true),
  data_id_(0),
  num_processed_(0),
  image_encoding_("rgb8")
{
  ParameterServer* ps = ParameterServer::instance();
  int q = ps->get<int>("subscriber_queue_size");
  if(ps->get<bool>("encoding_bgr")){
    image_encoding_ = "bgr8";//used in conversion to qimage. exact value does not matter, just not rgb8
  }
  std::string bagfile_name = ps->get<std::string>("bagfile_name");
  std::string visua_tpc = ps->get<std::string>("topic_image_mono");
  std::string depth_tpc = ps->get<std::string>("topic_image_depth");
  std::string cinfo_tpc = ps->get<std::string>("camera_info_topic");

  //Split-stuff
  std::string visua_tpc_split = ps->get<std::string>("topic_image_mono_split");
  std::string depth_tpc_split = ps->get<std::string>("topic_image_depth_split");
  std::string cinfo_tpc_split = ps->get<std::string>("camera_info_topic_split");

  ros::NodeHandle nh;
  tflistener_ = new tf::TransformListener(ros::Duration(60.0), true);
  if(bagfile_name.empty()){
    std::string cloud_tpc = ps->get<std::string>("topic_points");
    std::string widev_tpc = ps->get<std::string>("wide_topic");
    std::string widec_tpc = ps->get<std::string>("wide_cloud_topic");

    //All information from Kinect
    if(!visua_tpc.empty() && !depth_tpc.empty() && !cloud_tpc.empty())
    {   
        visua_sub_ = new image_sub_type(nh, visua_tpc, q);
        depth_sub_ = new image_sub_type (nh, depth_tpc, q);
        cloud_sub_ = new pc_sub_type (nh, cloud_tpc, q);  
        kinect_sync_ = new message_filters::Synchronizer<KinectSyncPolicy>(KinectSyncPolicy(q),  *visua_sub_, *depth_sub_, *cloud_sub_),
        kinect_sync_->registerCallback(boost::bind(&OpenNIListener::kinectCallback, this, _1, _2, _3));
        ROS_INFO_STREAM("Listening to " << visua_tpc << ", " << depth_tpc << " and " << cloud_tpc);
    } 
    else if(!visua_tpc_split.empty() && !depth_tpc_split.empty() && !cinfo_tpc_split.empty())
    {   
        visua_sub_ = new image_sub_type(nh, visua_tpc, q);
        depth_sub_ = new image_sub_type(nh, depth_tpc, q);
        cinfo_sub_ = new cinfo_sub_type(nh, cinfo_tpc, q);

        visua_sub_split_ = new image_sub_type(nh, visua_tpc_split, q);
        depth_sub_split_ = new image_sub_type(nh, depth_tpc_split, q);
        cinfo_sub_split_ = new cinfo_sub_type(nh, cinfo_tpc_split, q);

        no_cloud_sync_split_ = new message_filters::Synchronizer<SplitNoCloudSyncPolicy>(SplitNoCloudSyncPolicy(q),  *visua_sub_, *depth_sub_, *cinfo_sub_,  *visua_sub_split_, *depth_sub_split_, *cinfo_sub_split_);
        no_cloud_sync_split_->registerCallback(boost::bind(&OpenNIListener::noCloudCallbackSplit, this, _1, _2, _3, _4, _5, _6));
        ROS_INFO_STREAM("Listening to all kinds of topics, e.g."  << visua_tpc << " and " << depth_tpc << ", " << visua_tpc_split << " and " << depth_tpc_split);
    } 
    //No cloud, but visual image and depth
    else if(!visua_tpc.empty() && !depth_tpc.empty() && !cinfo_tpc.empty() && cloud_tpc.empty())
    {   
        visua_sub_ = new image_sub_type(nh, visua_tpc, q);
        depth_sub_ = new image_sub_type(nh, depth_tpc, q);
        cinfo_sub_ = new cinfo_sub_type(nh, cinfo_tpc, q);
        no_cloud_sync_ = new message_filters::Synchronizer<NoCloudSyncPolicy>(NoCloudSyncPolicy(q),  *visua_sub_, *depth_sub_, *cinfo_sub_);
        no_cloud_sync_->registerCallback(boost::bind(&OpenNIListener::noCloudCallback, this, _1, _2, _3, 0));
        ROS_INFO_STREAM("Listening to " << visua_tpc << " and " << depth_tpc);
    } 

    //All information from stereo                                               
    if(!widec_tpc.empty() && !widev_tpc.empty())
    {   
      visua_sub_ = new image_sub_type(nh, widev_tpc, q);
      cloud_sub_ = new pc_sub_type(nh, widec_tpc, q);
      stereo_sync_ = new message_filters::Synchronizer<StereoSyncPolicy>(StereoSyncPolicy(q), *visua_sub_, *cloud_sub_);
      stereo_sync_->registerCallback(boost::bind(&OpenNIListener::stereoCallback, this, _1, _2));
      ROS_INFO_STREAM("Listening to " << widev_tpc << " and " << widec_tpc );
    } 

    detector_ = createDetector(ps->get<std::string>("feature_detector_type"));
    extractor_ = createDescriptorExtractor(ps->get<std::string>("feature_extractor_type"));

    if(ps->get<bool>("concurrent_node_construction")){
      ROS_DEBUG("Threads used by QThreadPool on this Computer %i. Will increase this by one, b/c the QtRos Thread is very lightweight", QThread::idealThreadCount());
      //QThreadPool::globalInstance()->setMaxThreadCount(QThread::idealThreadCount()*2+2);
    }

    if(ps->get<bool>("use_robot_odom")){
    	odom_sub_= new odom_sub_type(nh, ps->get<std::string>("odometry_tpc"), 3);
    	odom_sub_->registerCallback(boost::bind(&OpenNIListener::odomCallback,this,_1));
    }
  } 
  detector_ = createDetector(ps->get<std::string>("feature_detector_type"));
  extractor_ = createDescriptorExtractor(ps->get<std::string>("feature_extractor_type"));
}

void OpenNIListener::evaluation(std::string filename)
{
    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "initial_guess");
    graph_mgr_->saveG2OGraph(QString(filename.c_str()) + "initial_guess.g2o");
    ROS_WARN_NAMED("eval", "Finished with optimization iteration initial_guess.");

    //INITIAL POSE GRAPH OPTIMIZATION
    ParameterServer::instance()->set<std::string>("pose_relative_to", std::string("first"));
    graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str()));//Non threaded call
    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "baseline");
    graph_mgr_->saveG2OGraph(QString(filename.c_str()) + "baseline.g2o");
    ROS_WARN_NAMED("eval", "Finished with optimization iteration baseline.");

    /*
    if(graph_mgr_->pruneEdgesWithErrorAbove(1, true) > 0){//Mahalanobis Distance
      graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str()));//Non threaded call
    } else {//if nothing has changed through pruning, only do one optimization iteration to get the same log output
      graph_mgr_->optimizeGraph(-1, true, QString(filename.c_str()));//Non threaded call
    }
    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "baselinePruned");
    ROS_WARN_NAMED("eval", "Finished with optimization iteration baselinePruned.");
    */


    graph_mgr_->setEmpiricalCovariances();
    graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str()));//Non threaded call

    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "empiricalCov");
    graph_mgr_->saveG2OGraph(QString(filename.c_str()) + "empiricalCov.g2o");
    ROS_WARN_NAMED("eval", "Finished with optimization iteration empiricalCov.");

    //Find calibration parameters...
    ParameterServer::instance()->set<std::string>("pose_relative_to", std::string("calibration"));
    graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str()));//Non threaded call
    //And optimize again
    ParameterServer::instance()->set<std::string>("pose_relative_to", std::string("first"));
    graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str()));//Non threaded call

    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "secEmpiricalCov");
    graph_mgr_->saveG2OGraph(QString(filename.c_str()) + "secEmpiricalCov.g2o");
    ROS_WARN_NAMED("eval", "Finished with optimization iteration secEmpiricalCov.");

    if(ParameterServer::instance()->get<bool>("store_pointclouds")){
      graph_mgr_->saveAllClouds(QString(filename.c_str()), true);//not threaded
    }

    if(!ParameterServer::instance()->get<bool>("use_gui")){
      Q_EMIT bagFinished();
      ROS_INFO("Waiting for all threads to finish.");
      usleep(10000000);//10sec to allow all threads to finish (don't know how much is required)
      ROS_INFO("Finished Waiting");
    }
    else{ 
      usleep(1000000);//1sec to allow output to flush
    }
  std::cerr << "Evaluation Done\n";
}
/* 
void OpenNIListener::evaluation(std::string filename)
{
    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "iteration_" + QString::number(0));
    ROS_WARN_NAMED("eval", "Finished with optimization iteration %i.", 0);

    //INITIAL POSE GRAPH OPTIMIZATION
    ParameterServer::instance()->set<std::string>("pose_relative_to", std::string("first"));
    graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str()));//Non threaded call
    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "iteration_" + QString::number(1));
    ROS_WARN_NAMED("eval", "Finished with optimization iteration %i.", 1);

    if(graph_mgr_->pruneEdgesWithErrorAbove(5) > 0){//Mahalanobis Distance
      graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str()));//Non threaded call
    } else {//if nothing has changed through pruning, only do one optimization iteration to get the same log output
      graph_mgr_->optimizeGraph(1, true, QString(filename.c_str()));//Non threaded call
    }
    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "iteration_" + QString::number(2));
    ROS_WARN_NAMED("eval", "Finished with optimization iteration %i.", 2);

    if(graph_mgr_->pruneEdgesWithErrorAbove(1) > 0){//Mahalanobis Distance
      graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str()));//Non threaded call
    } else {//if nothing has changed through pruning, only do one optimization iteration to get the same log output
      graph_mgr_->optimizeGraph(1, true, QString(filename.c_str()));//Non threaded call
    }
    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "iteration_" + QString::number(3));
    ROS_WARN_NAMED("eval", "Finished with optimization iteration %i.", 3);

    if(graph_mgr_->pruneEdgesWithErrorAbove(0.25) > 0){//Mahalanobis Distance
      graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str()));//Non threaded call
    } else {//if nothing has changed through pruning, only do one optimization iteration to get the same log output
      graph_mgr_->optimizeGraph(1, true, QString(filename.c_str()));//Non threaded call
    }
    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "iteration_" + QString::number(4));
    ROS_WARN_NAMED("eval", "Finished with optimization iteration %i.", 4);

    if(ParameterServer::instance()->get<bool>("optimize_landmarks")){ //LANDMARK OPTIMIZATION
      ParameterServer::instance()->set<bool>("optimize_landmarks", true);
      graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str())+"_landmark_optimized_");//Non threaded call
      graph_mgr_->saveTrajectory(QString(filename.c_str()) + "iteration_" + QString::number(3));
      ROS_WARN_NAMED("eval", "Finished with optimization iteration %i.", 3);
    }

    if(ParameterServer::instance()->get<bool>("octomap_online_creation")){
      graph_mgr_->writeOctomap(QString(filename.c_str()) + "-online.ot");
      //Now recreate (to have all clouds optimally positioned
      ParameterServer::instance()->set<bool>("octomap_online_creation", false);
      graph_mgr_->saveOctomap(QString(filename.c_str()) + "-offline.ot", false);
    }
    /*
    if(graph_mgr_->pruneEdgesWithErrorAbove(1) > 0){//Mahalanobis Distance
      graph_mgr_->optimizeGraph(-100, true, QString(filename.c_str()));//Non threaded call
    } else {//if nothing has changed through pruning, only do one optimization iteration to get the same log output
      graph_mgr_->optimizeGraph(1, true, QString(filename.c_str()));//Non threaded call
    }
    graph_mgr_->saveTrajectory(QString(filename.c_str()) + "iteration_" + QString::number(4));
    ROS_WARN_NAMED("eval", "Finished with optimization iteration %i.", 4);

    */
    /*
    if(ParameterServer::instance()->get<bool>("store_pointclouds")){
      graph_mgr_->saveIndividualClouds(QString(filename.c_str()), false);//not threaded
    }
    *//*

    if(!ParameterServer::instance()->get<bool>("use_gui")){
      Q_EMIT bagFinished();
      usleep(10000000);//10sec to allow all threads to finish (don't know how much is required)
    }
  std::cerr << "Evaluation Done\n";
}
*/



void OpenNIListener::displayOnly (const sensor_msgs::ImageConstPtr& visual_img_msg,
                                  int cam_id)
{
    ParameterServer* ps = ParameterServer::instance();
    if(ps->get<bool>("use_gui") && ps->get<bool>("show_2d_display"))
    {
      cv::Mat visual_img =  cv_bridge::toCvCopy(visual_img_msg)->image;
      image_encoding_ = visual_img_msg->encoding;
      Q_EMIT newVisualImage(cvMat2QImage(visual_img, 0+cam_id), cam_id); //visual_idx=0
    }
}


void OpenNIListener::displayOnly (const sensor_msgs::ImageConstPtr& visual_img_msg,
                                  const sensor_msgs::ImageConstPtr& depth_img_msg,
                                  int cam_id)
{
    ParameterServer* ps = ParameterServer::instance();
    if(ps->get<bool>("use_gui") && ps->get<bool>("show_2d_display"))
    {
      cv::Mat visual_img =  cv_bridge::toCvCopy(visual_img_msg)->image;
      cv::Mat depth_float_img = cv_bridge::toCvCopy(depth_img_msg)->image;
      if(visual_img.rows != depth_float_img.rows || 
         visual_img.cols != depth_float_img.cols){
        ROS_ERROR("depth and visual image differ in size! Ignoring Data");
        return;
      }
      image_encoding_ = visual_img_msg->encoding;
      Q_EMIT newVisualImage(cvMat2QImage(visual_img, 0+cam_id), cam_id); //visual_idx=0
      depthToCV8UC1(depth_float_img, depth_mono8_img_); //float can't be visualized or used as mask in float format TODO: reprogram keypoint detector to use float values with nan to mask
      Q_EMIT newDepthImage (cvMat2QImage(depth_mono8_img_,2+cam_id), cam_id);//overwrites last cvMat2QImage
    }
}



void OpenNIListener::mirrorCallback (const sensor_msgs::ImageConstPtr& visual_img_msg1,
                                     const sensor_msgs::ImageConstPtr& depth_img_msg1,
                                     const sensor_msgs::CameraInfoConstPtr& cam_info_msg1,
                                     const sensor_msgs::ImageConstPtr& visual_img_msg2,
                                     const sensor_msgs::ImageConstPtr& depth_img_msg2,
                                     const sensor_msgs::CameraInfoConstPtr& cam_info_msg2) 
{ 
  ScopedTimer s(__FUNCTION__);
  if(visual_img_msg1->header.frame_id.c_str() == visual_img_msg2->header.frame_id.c_str()){
    ROS_ERROR("Skipping image pair from same camera");
    return;
  }
  ROS_WARN_ONCE_NAMED("eval", "First RGBD-Data Received");
  ParameterServer* ps = ParameterServer::instance();
  if(++data_id_ < ps->get<int>("skip_first_n_frames")
     || data_id_ > ps->get<int>("stop_after_n_frames") 
     || (data_id_ % ps->get<int>("data_skip_step") != 0)
     || (pause_ && !getOneFrame_)
     || asyncFrameDrop(depth_img_msg1->header.stamp, visual_img_msg1->header.stamp)
     || asyncFrameDrop(depth_img_msg2->header.stamp, visual_img_msg2->header.stamp))
  { 
    displayOnly(visual_img_msg1, depth_img_msg1, 0);
    displayOnly(visual_img_msg2, depth_img_msg2, 1);
  } 
  else 
  {
    if(getOneFrame_) {
      getOneFrame_ = false;
    }
    Node* node1 = NULL;
    cv::Mat visual_img1, visual_img2;
    {//Processing first split
      ROS_WARN("Processing first part of split: %s", visual_img_msg1->header.frame_id.c_str());
      int cam_id = 0;
      cv::Mat depth_float_img;
      rosToOpenCV(visual_img_msg1, depth_img_msg1, visual_img1, depth_float_img);

      //VISUALIZATION
      if(ps->get<bool>("use_gui") && ps->get<bool>("show_2d_display")){
        Q_EMIT newVisualImage(cvMat2QImage(visual_img1, 0+cam_id), cam_id); //visual_idx=0
        depthToCV8UC1(depth_float_img, depth_mono8_img_); //float can't be visualized or used as mask in float format TODO: reprogram keypoint detector to use float values with nan to mask
        Q_EMIT newDepthImage (cvMat2QImage(depth_mono8_img_,2+cam_id), cam_id);//overwrites last cvMat2QImage
      }

      ScopedTimer s("Node creation for first split");
      node1 = new Node(visual_img1, depth_float_img, depth_mono8_img_, cam_info_msg1, depth_img_msg1->header, detector_, extractor_);
      retrieveTransformations(depth_img_msg1->header, node1);//Retrieve the transform between the lens and the base-link at capturing time;
    }
    Node* node2 = NULL;
    {
      ROS_WARN("Processing second part of split: %s", visual_img_msg2->header.frame_id.c_str());
      int cam_id = 1;
      cv::Mat depth_float_img;
      rosToOpenCV(visual_img_msg2, depth_img_msg2, visual_img2, depth_float_img);

      //VISUALIZATION
      if(ps->get<bool>("use_gui") && ps->get<bool>("show_2d_display")){
        Q_EMIT newVisualImage(cvMat2QImage(visual_img2, 0+cam_id), cam_id); //visual_idx=0
        depthToCV8UC1(depth_float_img, depth_mono8_img2_); //float can't be visualized or used as mask in float format TODO: reprogram keypoint detector to use float values with nan to mask
        Q_EMIT newDepthImage (cvMat2QImage(depth_mono8_img2_,2+cam_id), cam_id);//overwrites last cvMat2QImage
      }

      ScopedTimer s("Node creation for second split");
      node2 = new Node(visual_img2, depth_float_img, depth_mono8_img2_, cam_info_msg2, depth_img_msg2->header, detector_, extractor_);
      retrieveTransformations(depth_img_msg2->header, node2);//Retrieve the transform between the lens and the base-link at capturing time;
    }
    ROS_WARN("Merging Nodes");
    { //Merge nodes into node1
      g2o::Vector7d calibration_displacement; 
      //Mirror: 
      //6DOF Calib
      calibration_displacement << -0.0189617, 0.0707785, 0.104771, 0.938763, 0.00677368, -0.00478947, 0.344465;
      //calibration_displacement << -0.0110499, 0.0502848, 0.099574, 0.936359, 0.0142091, -0.00807437, 0.350664;
      //Measured Mirror (10cm from focal point to mirror)
      //calibration_displacement << 0.0, -0.064, 0.1766, 0.939692, 0.00677368, -0.00478947, 0.34202;
      //3DOF Calib
      //calibration_displacement << 0, 0.0547883,  0.107504,  0.936921,         0,        0,   0.34954;
      //2DOF Calib
      //calibration_displacement << 0, 0.000662044,  -0.0017672,    0.936444,           0,           0,    0.350818;
      //calibration_displacement << -0.0189617, 0.0707785, 0.104771, 0.938763, 0.00677368, -0.00478947, 0.344465;
      //Two Xtion Pro
      //calibration_displacement << 0.00157616, -4.08291e-05, -0.0600107, 4.88448e-05, 1, 2.30876e-05, 1.76034e-05;
      //Two Xtion Pro Measured:
      //calibration_displacement << 0.04, 0.0, -0.06, 0, 1, 0, 0;
      Eigen::Isometry3f cal_tf = g2o::internal::fromVectorQT(calibration_displacement).cast<float>();

      //Feature 3D positionen
      node1->feature_locations_3d_.reserve(node1->feature_locations_3d_.size() + node2->feature_locations_3d_.size());
      BOOST_FOREACH(const Eigen::Vector4f& position, node2->feature_locations_3d_){
        node1->feature_locations_3d_.push_back(cal_tf * position);
      }

      transformAndAppendPointCloud(*(node2->pc_col), *(node1->pc_col), cal_tf.matrix());

      ROS_WARN("Merging keypoints, %lu in View 1, %lu in View 2", node1->feature_locations_2d_.size(),  node2->feature_locations_2d_.size());
      node1->feature_locations_2d_.reserve(node1->feature_locations_2d_.size() + node2->feature_locations_2d_.size());
      BOOST_FOREACH(cv::KeyPoint kp, node2->feature_locations_2d_){
        kp.pt.y += visual_img1.rows; //only required for 2d display
        node1->feature_locations_2d_.push_back(kp);
      }

      cv::vconcat(node1->feature_descriptors_, node2->feature_descriptors_, /*destination:*/ node1->feature_descriptors_ );

      node1->feature_depth_stats_.insert(node1->feature_depth_stats_.end(), node2->feature_depth_stats_.begin(), node2->feature_depth_stats_.end());
      node1->feature_matching_stats_.insert(node1->feature_matching_stats_.end(), node2->feature_matching_stats_.begin(), node2->feature_matching_stats_.end());
      node1->siftgpu_descriptors.insert(node1->siftgpu_descriptors.end(), node2->siftgpu_descriptors.begin(), node2->siftgpu_descriptors.end());
    }
    delete node2; 
    ROS_WARN("Deleted node2");

    if(ps->get<bool>("use_gui") && ps->get<bool>("show_2d_display")){
      cv::vconcat(visual_img1, visual_img2, visual_img1);
    }
    callProcessing(visual_img1, node1, 0);
  }
}

void OpenNIListener::noCloudCallbackSplit (const sensor_msgs::ImageConstPtr& visual_img_msg1,
                                           const sensor_msgs::ImageConstPtr& depth_img_msg1,
                                           const sensor_msgs::CameraInfoConstPtr& cam_info_msg1,
                                           const sensor_msgs::ImageConstPtr& visual_img_msg2,
                                           const sensor_msgs::ImageConstPtr& depth_img_msg2,
                                           const sensor_msgs::CameraInfoConstPtr& cam_info_msg2) 
{ 
  if(visual_img_msg1->header.frame_id.c_str() == visual_img_msg2->header.frame_id.c_str()){
    ROS_ERROR("Skipping image pair");
    return;
  }
  //if(pause_ && !getOneFrame_) return;
  if(++data_id_ < ParameterServer::instance()->get<int>("skip_first_n_frames") / 2 ||
       data_id_ > ParameterServer::instance()->get<int>("stop_after_n_frames") )
  { 
    return;
  } else {
    bool two_frames = getOneFrame_; //always get both splits when one frame is requested
    int camera_id = 0;
    ROS_DEBUG("Processing first part of split: %s", visual_img_msg1->header.frame_id.c_str());
    noCloudCallback(visual_img_msg1, depth_img_msg1, cam_info_msg1, camera_id);
    camera_id = 1;
    getOneFrame_ = two_frames;
    ROS_DEBUG("Processing second part of split: %s", visual_img_msg2->header.frame_id.c_str());
    noCloudCallback(visual_img_msg2, depth_img_msg2, cam_info_msg2, camera_id);
  }
}

void OpenNIListener::rosToOpenCV(const sensor_msgs::ImageConstPtr& visual_img_msg,
                                 const sensor_msgs::ImageConstPtr& depth_img_msg,
                                 cv::Mat& visual_img,
                                 cv::Mat& depth_float_img)
{
  //IMAGE FORMAT CONVERSIONS
  depth_float_img = cv_bridge::toCvCopy(depth_img_msg)->image;
  if(image_encoding_ == "bayer_grbg8"){
    cv_bridge::toCvShare(visual_img_msg);
    ROS_INFO("Converting from Bayer to RGB");
    cv::cvtColor(cv_bridge::toCvCopy(visual_img_msg)->image, visual_img, CV_BayerGR2RGB, 3);
  } else{
    ROS_DEBUG_STREAM("Encoding: " << visual_img_msg->encoding);
    visual_img =  cv_bridge::toCvCopy(visual_img_msg)->image;
  }
  if(visual_img.rows != depth_float_img.rows || 
     visual_img.cols != depth_float_img.cols){
    ROS_ERROR("depth and visual image differ in size! Ignoring Data");
    return;
  }
  image_encoding_ = visual_img_msg->encoding;
}


void OpenNIListener::noCloudCallback (const sensor_msgs::ImageConstPtr& visual_img_msg,
                                      const sensor_msgs::ImageConstPtr& depth_img_msg,
                                      const sensor_msgs::CameraInfoConstPtr& cam_info_msg,
                                      int cam_id) 
{
  ScopedTimer s(__FUNCTION__);
  ROS_WARN_ONCE_NAMED("eval", "First RGBD-Data Received");
  ROS_DEBUG("Received data from kinect");

  //INITIAL SKIPPING CHECKS
  ParameterServer* ps = ParameterServer::instance();
  if(ps->get<std::string>("topic_image_mono_split") == "") ++data_id_;
  if(data_id_ < ps->get<int>("skip_first_n_frames") 
     || (data_id_ % ps->get<int>("data_skip_step") != 0)
     || (pause_ && !getOneFrame_)) 
  { 
    ROS_INFO_THROTTLE(5, "Skipping Frame %i because of data_skip_step setting (this msg is only shown every five seconds)", data_id_);
    displayOnly(visual_img_msg, depth_img_msg, cam_id);
    return;
  }
  if(asyncFrameDrop(depth_img_msg->header.stamp, visual_img_msg->header.stamp)) 
    return;

  cv::Mat visual_img, depth_float_img;
  rosToOpenCV(visual_img_msg, depth_img_msg, visual_img, depth_float_img);

  //VISUALIZATION
  if(ps->get<bool>("use_gui") && ps->get<bool>("show_2d_display")){
    Q_EMIT newVisualImage(cvMat2QImage(visual_img, 0+cam_id), cam_id); //visual_idx=0
    depthToCV8UC1(depth_float_img, depth_mono8_img_); //float can't be visualized or used as mask in float format TODO: reprogram keypoint detector to use float values with nan to mask
    Q_EMIT newDepthImage (cvMat2QImage(depth_mono8_img_,2+cam_id), cam_id);//overwrites last cvMat2QImage
  }

  noCloudCameraCallback(visual_img, depth_float_img, depth_mono8_img_, depth_img_msg->header, cam_info_msg, cam_id);
}





void OpenNIListener::cameraCallback(cv::Mat visual_img, 
                                    pointcloud_type::Ptr point_cloud, 
                                    cv::Mat depth_mono8_img)
{
  ScopedTimer s(__FUNCTION__);
  ROS_WARN_COND(point_cloud ==NULL, "Nullpointer for pointcloud");
  if(getOneFrame_) { getOneFrame_ = false; }//if getOneFrame_ is set, unset it and skip check for  pause
  else if(pause_) { return; }//Visualization and nothing else

  //######### Main Work: create new node ##############################################################
  Q_EMIT setGUIStatus("Computing Keypoints and Features");
  Node* node_ptr = new Node(visual_img, detector_, extractor_, point_cloud, depth_mono8_img);

  retrieveTransformations(point_cloud->header, node_ptr);
  callProcessing(visual_img, node_ptr, 0);
}






void OpenNIListener::noCloudCameraCallback(cv::Mat visual_img, 
                                           cv::Mat depth, 
                                           cv::Mat depth_mono8_img,
                                           std_msgs::Header depth_header,
                                           const sensor_msgs::CameraInfoConstPtr& cam_info,
                                           int cam_id)
{
  if(getOneFrame_) { //if getOneFrame_ is set, unset it and skip check for  pause
      getOneFrame_ = false;
  }
  ScopedTimer s(__FUNCTION__);
  Q_EMIT setGUIStatus("Computing Keypoints and Features");
  Node* node_ptr = new Node(visual_img, depth, depth_mono8_img, cam_info, depth_header, detector_, extractor_);

  retrieveTransformations(depth_header, node_ptr);//Retrieve the transform between the lens and the base-link at capturing time;

  callProcessing(visual_img, node_ptr, cam_id);
}




//Call function either regularly or as background thread
void OpenNIListener::callProcessing(cv::Mat visual_img, Node* node_ptr, int cam_id)
{
  if(!future_.isFinished()){
    ScopedTimer s("New Node is ready, waiting for graph manager.");
    future_.waitForFinished(); //Wait if GraphManager ist still computing. 
  }

  //update for visualization of the feature flow
  visualization_img_ = visual_img; //No copy
  visualization_depth_mono8_img_ = depth_mono8_img_;//No copy
  //With this define, processNode runs in the background and after finishing visualizes the results
  //Thus another Callback can be started in the meantime, to create a new node in parallel
  if(ParameterServer::instance()->get<bool>("concurrent_node_construction")) {
    ROS_DEBUG("Processing Node in parallel to the construction of the next node");
    if(ParameterServer::instance()->get<bool>("use_gui")){
      //visual_img points to the data received in the callback - which might be deallocated after the callback returns. 
      //This will happen before visualization if processNode is running in background, therefore the data needs to be cloned
      visualization_img_ = visual_img.clone();
      visualization_depth_mono8_img_ = depth_mono8_img_.clone();
    }
    future_ = QtConcurrent::run(this, &OpenNIListener::processNode, node_ptr, cam_id); 
  }
  else { //Non-concurrent
    processNode(node_ptr, cam_id);//regular function call
  }
}



void OpenNIListener::processNode(Node* new_node, int cam_id)
{
  ScopedTimer s(__FUNCTION__);
  ParameterServer* ps = ParameterServer::instance();

  Q_EMIT setGUIStatus("Adding Node to Graph");
  bool has_been_added = graph_mgr_->addNode(new_node);

  ///ODOMETRY
  if(has_been_added && !ps->get<std::string>("odom_frame_name").empty()){
    ros::Time latest_odom_time;
    std::string odom_frame  = ps->get<std::string>("odom_frame_name");
    std::string base_frame  = ps->get<std::string>("base_frame_name");
    std::string error_msg;
    int ev = tflistener_->getLatestCommonTime(odom_frame, base_frame, latest_odom_time, &error_msg); 
    if(ev == tf::NO_ERROR){
      graph_mgr_->addOdometry(latest_odom_time, tflistener_);
    } else {
      ROS_WARN_STREAM("Couldn't get time of lates tf transform between " << odom_frame << " and " << base_frame << ": " << error_msg);
    }
  }
  ++num_processed_;
  Q_EMIT setGUIInfo2(QString("Frames processed: ")+QString::number(num_processed_));

  //######### Visualization code  #############################################
  if(ps->get<bool>("use_gui") && ps->get<bool>("show_2d_display")){
    if(has_been_added){
      if(ps->get<bool>("visualize_mono_depth_overlay")){
        //Edge channels
        cv::Mat visual_edges = cv::Mat( visualization_img_.rows, visualization_img_.cols, CV_8UC1); 
        cv::Mat depth_edges  = cv::Mat( visualization_img_.rows, visualization_img_.cols, CV_8UC1); 
        overlay_edges(visualization_img_, depth_mono8_img_, visual_edges, depth_edges);

        //Feature channel
        cv::Mat feature_img = cv::Mat::zeros( visualization_img_.rows, visualization_img_.cols, CV_8UC1); 
        graph_mgr_->drawFeatureFlow(feature_img);

        //Q_EMIT newDepthImage(cvMat2QImage(depth_mono8_img_, 6), camera_id); //show registration
        Q_EMIT newFeatureFlowImage(cvMat2QImage(feature_img,depth_edges, visual_edges, 6), cam_id); //show registration by edge overlay
        //Q_EMIT newFeatureFlowImage(cvMat2QImage(visualization_img_,visualization_depth_mono8_img_, feature_img, 2), camera_id); //show registration by color overlay
      } else {
        graph_mgr_->drawFeatureFlow(visualization_img_, cv::Scalar(0,0,255), cv::Scalar(0,128,0) );
        //graph_mgr_->drawFeatureFlow(depth_mono8_img_, cv::Scalar(0,0,255), cv::Scalar(0,128,0) );
        Q_EMIT newFeatureFlowImage(cvMat2QImage(visualization_img_, 5+cam_id), cam_id); //show registration
        //Q_EMIT newDepthImage(cvMat2QImage(depth_mono8_img_, 7+cam_id), cam_id); //show registration
      }
    } else {
      if(ps->get<bool>("visualize_mono_depth_overlay")){
        cv::Mat feature_img = cv::Mat( visualization_img_.rows, visualization_img_.cols, CV_8UC1); 
        cv::drawKeypoints(feature_img, new_node->feature_locations_2d_, feature_img, cv::Scalar(155), 5);
        Q_EMIT newFeatureFlowImage(cvMat2QImage(visualization_img_,visualization_depth_mono8_img_, feature_img, 2), cam_id); //show registration
      } else {
        cv::drawKeypoints(visualization_img_, new_node->feature_locations_2d_, visualization_img_, cv::Scalar(0, 100,0), 5);
        Q_EMIT newFeatureFlowImage(cvMat2QImage(visualization_img_, 2), cam_id); //show registration
      }
    }
  }
  if(!has_been_added) {
    delete new_node; 
    new_node = NULL;
  }
}



//! Callback for the robot odometry
void OpenNIListener::odomCallback(const nav_msgs::OdometryConstPtr& odom_msg){
  tf::Transform tfTransf;
  tf::poseMsgToTF (odom_msg->pose.pose, tfTransf);
  tf::StampedTransform stTransf(tfTransf, odom_msg->header.stamp, odom_msg->header.frame_id, odom_msg->child_frame_id); 
  printTransform("Odometry Transformation", stTransf);
  tflistener_->setTransform(stTransf);
  //Is done now after creation of Node
	//graph_mgr_->addOdometry(odom_msg->header.stamp,tflistener_);
}

